#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static char     PCchars[] = {
        ' ',    /* Empty Space       */
        '',    /* Normal Floor      */
        '',    /* Solid Floor       */
        '',    /* Normal Ladder     */
        '',    /* Horizontal Rope   */
        '',    /* Fake Floor        */
        '|',    /* Escape Ladder     */
        '',    /* Gold              */
        '',    /* Enemy             */
        ''     /* Player            */
};

static char     textchars[] = {
        ' ',    /* Empty Space       */
        '-',    /* Normal Floor      */
        '=',    /* Solid Floor       */
        '#',    /* Normal Ladder     */
        '\"',    /* Horizontal Rope   */
        '~',    /* Fake Floor        */
        '|',    /* Escape Ladder     */
        'o',    /* Gold              */
        'E',    /* Enemy             */
        'P'     /* Player            */
};




static unsigned char    filebuf[0x200];
static unsigned char    lvlmap[26][16];
static char             lvlname[11];
static int guypos [6][2];
static int sizex, sizey;
static int unknown1, gold, unknown2, numenemy;



static unsigned char pcdecrypt[] =
        { 0xE3, 0xB2, 0x71, 0xD9, 0xB8, 0xEC, 0x5C, 0x76,
          0x2E, 0x3B, 0x97, 0x1D, 0xCB, 0x8E, 0x65, 0xC7,
          0xB2, 0xE3, 0xD9, 0x71, 0xEC, 0xB8, 0x76, 0x5C,
          0x3B, 0x2E, 0x1D, 0x97, 0x8E, 0xCB, 0xC7, 0x65 };

/*  This precalculated array is generated by taking the 16-bit value 0xE3B2
    and rotating it right by one bit through all 16 positions.  This array
    is then XORed with the data to encrypt or decrypt. */

int loaddatv2(char *filename, int levelnum)
/* Load a level from a PC version 2+ file (loderun.dat) */
{
        FILE    *f;
        int     i, j, k;
        
        /* snork in a level from the file */
        f = fopen(filename,"rb");
        if (fseek(f, (levelnum - 1) * 0x200, SEEK_SET)) return 2;
        if (fread(filebuf, 1, 0x200, f) < 0x200) return 3;
        fclose(f);

        /* decrypt the level */
        for (i = 0; i < 0x200; i++) {
            filebuf[i] ^= pcdecrypt[i & 0x1f];
        }


        /* copy the level map */
        for (i = 0, k = 0; i < 26; i++) {
                for (j = 0; j < 16; j++, k++) {
                        lvlmap[i][j] = filebuf[k];
                }
        }

        /* copy some level parameters */
        sizex = 26;
        sizey = 16;
        unknown1 = (int) filebuf[0x1a0];   /* Seems to always be 0xFF */
        gold     = (int) filebuf[0x1a1];   /* Number of gold to finish */
        unknown2 = (int) filebuf[0x1a2];   /* Seems to always be 0x01 */
        numenemy = (int) filebuf[0x1a3];   /* Number of enemies */

        /* copy the guy positions */
        for (i = 0, k = 0x1a4; i < 6; i++) {
                for (j = 0; j < 2; j++, k++) {
                        guypos[i][j] = (int) filebuf[k];                        
                }
        }

        /* copy the level name */
        strcpy(lvlname, &filebuf[0x1f5]);


        return 0;
}

int savedatv2(char *filename, int levelnum)
/* Save a level to a PC version 2+ file (loderun.dat) */
{
        FILE    *f;
        int     i, j, k;
        
        /* clear the filebuffer */
        memset(filebuf, 0, sizeof(filebuf));

        /* copy the level map */
        for (i = 0, k = 0; i < 26; i++) {
                for (j = 0; j < 16; j++, k++) {
                         filebuf[k] = lvlmap[i][j];
                }
        }

        /* copy some level parameters */
        filebuf[0x1a0] = (unsigned char) unknown1;   /* Seems to always be 0xFF */
        filebuf[0x1a1] = (unsigned char) gold;       /* Number of gold to finish */
        filebuf[0x1a2] = (unsigned char) unknown2;   /* Seems to always be 0x01 */
        filebuf[0x1a3] = (unsigned char) numenemy;   /* Number of enemies */


        /* copy the guy positions */
        for (i = 0, k = 0x1a4; i < 6; i++) {
                for (j = 0; j < 2; j++, k++) {
                        filebuf[k] = (unsigned char) guypos[i][j];                         
                }
        }



        /* insert loderunner string */
        strcpy(&filebuf[0x1f5], lvlname);

        /* encrypt the level */
        for (i = 0; i < 0x200; i++) {
            filebuf[i] ^= pcdecrypt[i & 0x1f];
        }

        /* save the level to the file */
        f = fopen(filename,"ab");
        if (fseek(f, (levelnum - 1) * 0x200, SEEK_SET)) return 2;
        if (fwrite(filebuf, 1, 0x200, f) < 0x200) return 3;
        fclose(f);

        return 0;
}



void printlevel(FILE *f)
{
        int     x, y;
        for (y = 0; y < sizey; y++) {
                for (x = 0; x <sizex; x++) {
                        fputc (PCchars[lvlmap[x][y]],f);
                }
                fputc ('\n', f);
        }
}





int parselrl(char *s)
/* parse a line from an .lrl file */
/* return:     */
/* 0 = success */
/* 1 = missing parameter */
         

{
        char p[256];
        int i,j,k;
        if (isdigit(s[0]) && isdigit(s[1])) {
                j = (s[0] - '0') * 10 + (s[1] - '0');
                for (i = 0; (i < sizex + 0) && (s[i + 3] != 0); i++) {
                        for (k = 0; k < sizeof(textchars); k++) {
                                if (s[i + 3] == textchars[k])
                                        lvlmap[i][j] = k;
                        }
                }
                return 0;
        }

        if (sscanf(s, "%s", p) == 1) {
                for (i = 0; p[i] != 0; i++) p[i] = tolower(p[i]);
                
                if (strcmp(p, "sizex") == 0) {
                        if (sscanf(s, "%s %i", p, &sizex) < 2) return 1;
                        return 0;
                }
                if (strcmp(p, "sizey") == 0) {
                        if (sscanf(s, "%s %i", p, &sizey) < 2) return 1;
                        return 0;
                }
                if (strcmp(p, "unknown1") == 0) {
                        if (sscanf(s, "%s %i", p, &unknown1) < 2) return 1;
                        return 0;
                }
                if (strcmp(p, "gold") == 0) {
                        if (sscanf(s, "%s %i", p, &gold) < 2) return 1;
                        return 0;
                }
                if (strcmp(p, "unknown2") == 0) {
                        if (sscanf(s, "%s %i", p, &unknown2) < 2) return 1;
                        return 0;
                }
                if (strcmp(p, "numenemy") == 0) {
                        if (sscanf(s, "%s %i", p, &numenemy) < 2) return 1;
                        return 0;
                }
                
                if (strcmp(p, "player") == 0) {
                        if (sscanf(s, "%s ( %i , %i )", p, &guypos[0][0], &guypos[0][1]) < 3)
                                return 1;
                        return 0;
                }
                if (strcmp(p, "enemy1") == 0) {
                        if (sscanf(s, "%s ( %i , %i )", p, &guypos[1][0],&guypos[1][1]) < 3)
                                return 1;
                        return 0;
                }
                if (strcmp(p, "enemy2") == 0) {
                        if (sscanf(s, "%s ( %i , %i )", p, &guypos[2][0],&guypos[2][1]) < 3)
                                return 1;
                        return 0;
                }
                if (strcmp(p, "enemy3") == 0) {
                        if (sscanf(s, "%s ( %i , %i )", p, &guypos[3][0],&guypos[3][1]) < 3)
                                return 1;
                        return 0;
                }
                if (strcmp(p, "enemy4") == 0) {
                        if (sscanf(s, "%s ( %i , %i )", p, &guypos[4][0],&guypos[4][1]) < 3)
                                return 1;
                        return 0;
                }
                if (strcmp(p, "enemy5") == 0) {
                        if (sscanf(s, "%s ( %i , %i )", p, &guypos[5][0],&guypos[5][1]) < 3)
                                return 1;
                        return 0;
                }
                if (strcmp(p, "lvlname") == 0) {
                        if (sscanf(s, "%s %s", p, lvlname) < 2) return 1;
                        return 0;
                }                       
        }
        return 2;

}



int loadlrl(char *filename)
/* load a human-readable .lrl file */
{
        FILE    *f;
        int     line, err;
        char    s[256];    

        /* put in some sane defaults */
        sizex = 26;
        sizey = 16;
        gold = 1;
        numenemy = 0;
        strcpy(lvlname, "LodeRunner");

        /* open the file */
        f = fopen(filename,"r");
        for (line = 1; fgets(s,255,f) != NULL; line++) {
                if (s[0] != '#') err = parselrl(s);
                if (err == 1) {
                        printf("Line %i: Expected Parameter Missing", line);
                }
        }                
        /* close the file */
        fclose(f);



}



int savelrl(char *filename)
/* save a human-readable .lrl file */
{
        int     x, y;
        FILE    *f;

        /* open the file */
        f = fopen(filename,"w");
        if (fprintf(f, "# Lode Runner Level File\n") < 0) return 1;
        fputs("\n",f);
        fprintf(f, "# Author: Generated by LODETOOL\n");
        fputs("\n",f);

        /* write the level name */
        fprintf(f, "# Level name (\"LodeRunner\" in standard file, seems you can replace it\n");
        fprintf(f, "# with your own string, up to a max of 10 chars.\n");
        fprintf(f, "lvlname         %s\n",lvlname);
        fputs("\n",f);


        /* write parameters */
        fprintf(f, "# Unknown value 1, seems to be always 255.\n");
        fprintf(f, "unknown1        %i\n",unknown1);
        fputs("\n",f);

        fprintf(f, "# Number of gold needed to complete level.  May be less than or equal to\n");
        fprintf(f, "# amount present in level.\n");
        fprintf(f, "gold            %i\n",gold);
        fputs("\n",f);

        fprintf(f, "# Unknown value 2, seems to be always 1.\n");
        fprintf(f, "unknown2        %i\n",unknown2);
        fputs("\n",f);

        fprintf(f, "# Player start location, (x,y) where upper left is (0,0)\n");
        fprintf(f, "player          (%i,%i)\n",guypos[0][0],guypos[0][1]);
        fputs("\n",f);

        /* write enemy parameters */
        fprintf(f, "# Number of enemies, and enemy start locations.  Positions for unused\n");
        fprintf(f, "# enemies may be any value, game defaults to 255.  Location is (x,y) \n");
        fprintf(f, "# where upper left is (0,0).  Max number of enemies is 5 in PC version.\n");
        fprintf(f, "numenemy        %i\n",numenemy);
        fprintf(f, "enemy1          (%i,%i)\n",guypos[1][0],guypos[1][1]);
        fprintf(f, "enemy2          (%i,%i)\n",guypos[2][0],guypos[2][1]);
        fprintf(f, "enemy3          (%i,%i)\n",guypos[3][0],guypos[3][1]);
        fprintf(f, "enemy4          (%i,%i)\n",guypos[4][0],guypos[4][1]);
        fprintf(f, "enemy5          (%i,%i)\n",guypos[5][0],guypos[5][1]);
        fputs("\n",f);




        /* dump the level map */
        fputs("# Level map, in human readable format.  Editable too, if you're\n",f);
        fputs("# careful to keep everything in the same spot, and your editor doesn't\n",f);
        fputs("# insert tabs for spaces on save.  Hint: Use 50 line VGA mode\n",f);
        fputs("# for better aspect ratio.  Maps can be created that the editor\n",f);
        fputs("# won't allow, but are playable with interesting results.\n",f);
        fputs("\n",f);
        fputs("#Size of map\n",f);
        fprintf(f, "sizex           %i\n",sizex);
        fprintf(f, "sizey           %i\n",sizey);
        fputs("\n",f);
        fputs("#  0         1         2     \n",f);
        fputs("#  01234567890123456789012345\n",f);
        for (y = 0; y < sizey; y++) {
                fprintf (f, "%02i ",y);
                for (x = 0; x <sizex; x++) {
                        fputc (textchars[lvlmap[x][y]],f);
                }
                fputc ('\n', f);
        }
        fputs("#  01234567890123456789012345\n",f);
        fputs("#  0         1         2     \n#\n",f);
        fputs("# Legend:\n#",f);
        fprintf(f," Empty Space:     '%c' ",textchars[0]);
        fprintf(f," Normal Floor:     %c  ",textchars[1]);
        fprintf(f," Solid Floor:      %c  \n#",textchars[2]);
        fprintf(f," Normal Ladder:    %c  ",textchars[3]);
        fprintf(f," Horizontal Rope:  %c  ",textchars[4]);
        fprintf(f," Fake Floor:       %c  \n#",textchars[5]);
        fprintf(f," Escape Ladder:    %c  ",textchars[6]);
        fprintf(f," Gold:             %c  ",textchars[7]);
        fprintf(f," Enemy:            %c  \n#",textchars[8]);
        fprintf(f," Player:           %c  \n\n",textchars[9]);

        /* close the file */
        fclose(f);

        return 0;
}


int flen(char *filename)
/* open a file and return its length or -1 if not exist */
{
        FILE    *f;
        int     i;
        f = fopen   (filename,"r");
        if (fseek(f,0,SEEK_END)) {
                fclose(f);
                return -1;
        }
        i = ftell(f);
        fclose(f);
        return i;
}



int main(int argc, char *argv[])
{
        int   ilvl, olvl, err, a;
        char  p[256];
  
        fprintf(stderr, "LODETOOL version 0.1, maintained by Joshua Neal.\n");
        fprintf(stderr, "Copyright (C) 1999, Joshua Neal\n");
        fprintf(stderr, "This is free software; see the source for copying conditions.\n");
        fprintf(stderr, "There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A\n");
        fprintf(stderr, "PARTICULAR PURPOSE.\n\n");
        fprintf(stderr, "Report bugs to <tinara@goodnet.com>.\n\n");


        if (strcmp(argv[1],"-c") == 0) {
        /* -c file1 lvl file2 lvl */
                if (argc < 3) {
                        printf ("lodetool -c inputfile [levelnum] outputfile [levelnum]\n");
                        return 1;
                }
                a = 2;

                if (strstr(argv[a],".lrl")) {
                        err = loadlrl(argv[a]);
                        a++;
                } else if (strstr(argv[a],".dat")) {
                        err = loaddatv2(argv[a], atoi(argv[a + 1]));
                        a += 2;
                }


                        
                if (strstr(argv[a],".lrl")) {
                        err = savelrl(argv[a]);
                        a++;
                } else if (strstr(argv[a],".dat")) {
                        err = savedatv2(argv[a], atoi(argv[a + 1]));
                        a += 2;
                }
                
        }
        if (strcmp(argv[1],"-p") == 0) {
                if (argc < 2) {
                        printf ("lodetool -p inputfile [levelnum]\n");
                        return 1;
                }
                a = 2;

                if (strstr(argv[a],".lrl")) {
                        err = loadlrl(argv[a]);
                        a++;
                } else if (strstr(argv[a],".dat")) {
                        err = loaddatv2(argv[a], atoi(argv[a + 1]));
                        a += 2;
                }
                printlevel(stdout);
        }


}

